/* parser.c -- Parsing a scenario */

#include <stdio.h>
#include <ctype.h>
#include "locksim.h"

static int lookahead;

static char *tokentype2string(int t) {
	if (t == TRANSACTION)
		return "Transaction";
	else if (t == READ)
		return "Read";
	else if (t == WRITE)
		return "Write";
	else if (t == ID)
		return "identifier";
	else if (t == NUMBER)
		return "number";
	else if (t == DONE)
		return "DONE";
	else if (t == LOCK)
		return "LOCK";
	else if (t == READLOCK)
		return "ReadLock";
	else if (t == WRITELOCK)
		return "WriteLock";
	else if (t == UNLOCK)
		return "Unlock";
	else if (isprint(t)) {
		/* Mad hack so it works with "match" below */
		static int state = 0;
		static char s[2][2];
		if (++state > 1)
			state = 0;
		s[state][0] = t;
		s[state][1] = '\0';
		return s[state];
	}
	else
		return "something strange";
}

static void match(int expected, FILE *input) {
	if (lookahead != expected)
		fatal_error("Syntax error on line %d. Expected %s but got %s",
		line_number, tokentype2string(expected), tokentype2string(lookahead));
	else
		lookahead = scan(input);
}

static struct TreeNode *scenario(FILE *input),
*transactions(FILE *input),
*transaction(FILE *input),
*anycommands(FILE *input),
*anycommand(FILE *input),
*setcommands(FILE *input),
*setcommand(FILE *input),
*dbcommand(FILE *input),
*lockcommand(FILE *input),
*expression(FILE *input),
*term(FILE *input),
*factor(FILE *input);

struct TreeNode *parse_scenario(FILE *input) {
	lookahead = scan(input);
	return scenario(input);
}

static struct TreeNode *scenario(FILE *input) {
	struct TreeNode *result = NULL;
	struct TreeNode *cs = setcommands(input);
	struct TreeNode *ts = transactions(input);
	result = create_node(';', cs, ts);
	// match(DONE, input); -- No, reads another token!
	if (lookahead != DONE)
		fatal_error("Syntax error on line %d: Unexpected %s",
		line_number, tokentype2string(lookahead));
	return result;
}

static struct TreeNode *transactions(FILE *input) {
	struct TreeNode *ts = NULL;
	struct TreeNode *last_node = NULL;
	while (lookahead == TRANSACTION) {
		struct TreeNode *t = transaction(input);
		if (ts == NULL)
			ts = last_node = create_node(';', t, NULL);
		else {
			struct TreeNode *new_last_node = create_node(';', t, NULL);
			rplacd(last_node, new_last_node);
			last_node = new_last_node;
		}
	}
	return ts;
}

static struct TreeNode *transaction(FILE *input) {
	match(TRANSACTION, input);
	int symbol_number = lexical_value;
	match(ID, input);
	match('{', input);
	struct TreeNode *cs = anycommands(input);
	match('}', input);
	return create_node(TRANSACTION, create_id_node(symbol_number), cs);
}

static struct TreeNode *anycommands(FILE *input) {
	struct TreeNode *cs = NULL;
	struct TreeNode *last_node = NULL;
	while (lookahead == ID
		|| lookahead == READ || lookahead == WRITE
		|| lookahead == LOCK || lookahead == READLOCK || lookahead == WRITELOCK || lookahead == UNLOCK) {
		struct TreeNode *c = anycommand(input);
		if (cs == NULL)
			cs = last_node = create_node(';', c, NULL);
		else {
			struct TreeNode *new_last_node = create_node(';', c, NULL);
			rplacd(last_node, new_last_node);
			last_node = new_last_node;
		}
	}
	return cs;
}

static struct TreeNode *anycommand(FILE *input) {
	struct TreeNode *result = NULL;
	if (lookahead == ID)
		result = setcommand(input);
	else if (lookahead == READ || lookahead == WRITE)
		result = dbcommand(input);
	else if (lookahead == LOCK || lookahead == READLOCK || lookahead == WRITELOCK || lookahead == UNLOCK)
		result = lockcommand(input);
	else
		fatal_error("Syntax error on line %d: Unexpected %s",
		line_number, tokentype2string(lookahead));
	return result;
}

static struct TreeNode *setcommands(FILE *input) {
	struct TreeNode *cs = NULL;
	struct TreeNode *last_node = NULL;
	while (lookahead == ID) {
		struct TreeNode *c = setcommand(input);
		if (cs == NULL)
			cs = last_node = create_node(';', c, NULL);
		else {
			struct TreeNode *new_last_node = create_node(';', c, NULL);
			rplacd(last_node, new_last_node);
			last_node = new_last_node;
		}
	}
	return cs;
}

static struct TreeNode *setcommand(FILE *input) {
	int symbol_number = lexical_value;
	match(ID, input);
	match('=', input);
	struct TreeNode *e = expression(input);
	match(';', input);
	return create_node('=', create_id_node(symbol_number), e);
}

static struct TreeNode *dbcommand(FILE *input) {
	struct TreeNode *result = NULL;
	if (lookahead == READ || lookahead == WRITE) {
		int op = lookahead;
		match(op, input);
		match('(', input);
		int symbol_number = lexical_value;
		match(ID, input);
		match(')', input);
		match(';', input);
		result = create_node(op, create_id_node(symbol_number), NULL);
	}
	else
		fatal_error("Syntax error on line %d: Unexpected %s",
		line_number, tokentype2string(lookahead));
	return result;
}

static struct TreeNode *lockcommand(FILE *input) {
	struct TreeNode *result = NULL;
	if (lookahead == LOCK || lookahead == READLOCK || lookahead == WRITELOCK || lookahead == UNLOCK) {
		int op = lookahead;
		match(op, input);
		match('(', input);
		int symbol_number = lexical_value;
		match(ID, input);
		match(')', input);
		match(';', input);
		result = create_node(op, create_id_node(symbol_number), NULL);
	}
	else
		fatal_error("Syntax error on line %d: Unexpected %s",
		line_number, tokentype2string(lookahead));
	return result;
}

static struct TreeNode *expression(FILE *input) {
	struct TreeNode *e = term(input);
	while (lookahead == '+' || lookahead == '-') {
		int op = lookahead;
		match(op, input);
		struct TreeNode *t = term(input);
		e = create_node(op, e, t);
	}
	return e;
}

static struct TreeNode *term(FILE *input) {
	struct TreeNode *t = factor(input);
	while (lookahead == '*' || lookahead == '/' || lookahead == '%') {
		int op = lookahead;
		match(op, input);
		struct TreeNode *f = factor(input);
		t = create_node(op, t, f);
	}
	return t;
}

static struct TreeNode *factor(FILE *input) {
	struct TreeNode *result = NULL;
	if (lookahead == NUMBER) {
		int number = lexical_value;
		match(NUMBER, input);
		result = create_number_node(number);
	}
	else if (lookahead == ID) {
		int symbol_number = lexical_value;
		match(ID, input);
		result = create_id_node(symbol_number);
	}
	else if (lookahead == '(') {
		match('(', input);
		result = expression(input);
		match(')', input);
	}
	else
		fatal_error("Syntax error on line %d: Unexpected %s",
		line_number, tokentype2string(lookahead));

	return result;
}
